/*->c.tfile */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "transform.h"
#include "xprocess.h"
#include "deb.h"


#include "err.h"
#include "fsx.h"
#include "task.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "etc.h"
#include "bf.h"
#include "dbhi.h"
#include "key.h"
#include "config.h"
#include "pane.h"

#include "str.h"

#include "im.h"
#include "main.h"


#define USE_CONST 1
#define USE_PROTOTYPES 1

#include "tiffio.h"

#include "tload.h"
#include "tfile.h"



/*****************************************************************************/


static int tdbl;
static int dbl;
static int tcolour=COMPRESSION_LZW;
static int colour=COMPRESSION_LZW;
static int tbw=COMPRESSION_CCITTFAX3;
static int bw=COMPRESSION_CCITTFAX3;


static int givewarnings=0;


static char tifferrorstring[256];

static void errorhandler(const char * module,const char * fmt,va_list ap)
{
 int len;


 if(module!=NULL) len=sprintf(tifferrorstring, "%s: ", module);
 else             len=0;

 vsprintf(tifferrorstring+len,fmt,ap);


/* dprintf(1,"error handler called %s",tifferrorstring); */
}


void tiffclearerror(void)
{
 *tifferrorstring=0;
}


os_error * tiffgeterror(void)
{
 if(*tifferrorstring) return(generror(0,tifferrorstring));
 else                 return(NULL);
}


static void warninghandler(const char * module,const char * fmt,va_list ap)
{
 int  len;
	char warning[256];

	if(givewarnings)
	{
  if(module!=NULL) len=sprintf(warning, "%s: ", module);
  else             len=0;

  vsprintf(warning+len,fmt,ap);

// 	messagebox(warning);
	}
}




static os_error * tiff_loadframe(TIFF * tif,filestr * file,int fn)
{
 os_error * err;
 imagestr * loadimage;
 imagestr * maskimage;


 err=ximnew2(file->frames[fn].fxpix,file->frames[fn].fypix,
            file->frames[fn].fbpp,file->frames[fn].mbpp,&file->frames[fn].xim);

// cprintf("tiffloadframe bpp %d mbpp %d",file->frames[fn].fbpp,file->frames[fn].mbpp);

 if(!err)
 {
  loadimage=((file->frames[fn].xim)->sim[IM]);
  maskimage=((file->frames[fn].xim)->sim[AL]);

  loadimage->xdpi=file->frames[fn].fxdpi;
  loadimage->ydpi=file->frames[fn].fydpi;
  if(maskimage)
  {
   maskimage->xdpi=file->frames[fn].fxdpi;
   maskimage->ydpi=file->frames[fn].fydpi;
  }

//  getinfo(&file->frames[fn],tif);

  err=tiff_loadimage(tif,loadimage,maskimage,file->frames[fn].fxpix,file->frames[fn].fypix);
 }

 return(err);
}



os_error * loadtiff(char * name,int type,int userhandle,int xvolatile)
{
 os_error  * err;
 int         first;
 int         fn;
 TIFF      * tif;
 unsigned short bitspersample;
 unsigned short samplesperpixel;
 unsigned short   * sampleinfo;
 unsigned short     extrasamples;
 float       xres;
 float       yres;
 int         gotx;
 int         goty;
 unsigned short units;
 unsigned short photometric;

// cprintf("loadtiff");


 TIFFSetErrorHandler(errorhandler);
 TIFFSetWarningHandler(warninghandler);

 tiffclearerror();

// cprintf("clear error++");

 err=NULL;
 first=1;

 tif=TIFFOpen(name,"r");
 if(tif!=NULL)
 {
//  cprintf("TIFF open");

  do
  {
   err=addframe(loadfile,&fn,0);
   if(!err)
   {
    TIFFGetField(tif,TIFFTAG_IMAGEWIDTH,&loadfile->frames[fn].fxpix);
    TIFFGetField(tif,TIFFTAG_IMAGELENGTH,&loadfile->frames[fn].fypix);
    TIFFGetFieldDefaulted(tif,TIFFTAG_BITSPERSAMPLE,&bitspersample);
    TIFFGetFieldDefaulted(tif,TIFFTAG_SAMPLESPERPIXEL,&samplesperpixel);
    TIFFGetFieldDefaulted(tif,TIFFTAG_EXTRASAMPLES,&extrasamples,&sampleinfo);

    if(!TIFFGetField(tif,TIFFTAG_PHOTOMETRIC,&photometric))
    {
     switch(samplesperpixel-extrasamples)
     {
      case 1:
             photometric=PHOTOMETRIC_MINISBLACK;
             break;

      case 3: 
      case 4:
             photometric=PHOTOMETRIC_RGB;
             break;


     default:
  //  TIFFError(TIFFFileName(tif),"Missing needed \"PhotometricInterpretation\" tag");
             break;
     }
    }



//   cprintf("bitspersample=%d samplesperpixel=%d extra %d",bitspersample,
//                                         samplesperpixel,extrasamples);

    samplesperpixel=samplesperpixel-extrasamples;
    if(photometric==PHOTOMETRIC_RGB && samplesperpixel==4) 
    {
     extrasamples++;
     samplesperpixel--;
    }

    if(photometric==PHOTOMETRIC_SEPARATED && samplesperpixel==5) 
    {
     extrasamples++;
     samplesperpixel--;
    }

    if(samplesperpixel>4) samplesperpixel=4;
    if(extrasamples>4)    extrasamples=4;

    sprintf(loadfile->frames[fn].name,"TIFF%d",fn);
    loadfile->frames[fn].offset=0;
    loadfile->frames[fn].fbpp=bitspersample*samplesperpixel;
    loadfile->frames[fn].mbpp=bitspersample*extrasamples;


    xres=yres=90;

    if(!TIFFGetField(tif,TIFFTAG_RESOLUTIONUNIT,&units)) units=RESUNIT_INCH;

    {

    /* dprintf(0,"point 1"); */

     gotx=TIFFGetField(tif,TIFFTAG_XRESOLUTION,&xres);
     goty=TIFFGetField(tif,TIFFTAG_YRESOLUTION,&yres);

     if(gotx || goty)
     {
      if(!gotx) xres=yres;
      if(!goty) yres=xres;

      if(!xres) xres=90;
      if(!yres) yres=90;

     /* dprintf(0,"point 2 x=%f y=%f units=%d",xres,yres,units);  */

      switch(units)
      {
             case RESUNIT_NONE:
                               yres=(90/xres)*yres;
                               xres=90;
                               break;

             case RESUNIT_INCH:
                               break;

       case RESUNIT_CENTIMETER:
                               xres*=2.54;
                               yres*=2.54;
                               break;

                       default:
                               yres=(90/xres)*yres;
                               xres=90;
                               break;
      }
     }
    }

  /*  dprintf(2,"unit=%d xres=%g yres=%g",units,xres,yres);
    dprintf(3,"xpix=%d ypix=%d",loadfile->frames[fn].xpix,
                                loadfile->frames[fn].ypix); */

    loadfile->frames[fn].fxdpi=(int)xres;
    loadfile->frames[fn].fydpi=(int)yres;

    loadfile->frames[fn].xim=NULL;

    if(first)
    {
     err=tiff_loadframe(tif,loadfile,fn);
     first=0;
    }
   }
  } while(TIFFReadDirectory(tif));

  TIFFClose(tif);
 }

 if(!err) err=tiffgeterror();

 return(err);

 USE(type);
 USE(userhandle);
 USE(xvolatile);
}




static os_error * getframe(filestr * file,int fn)
{
 os_error  * err;
 TIFF      * tif;
 int         count;

 tiffclearerror();

 err=NULL;
 count=0;



 tif=TIFFOpen(file->oname,"r");
 if(tif!=NULL)
 {
  while(count<fn)
  {
   if(!TIFFReadDirectory(tif)) break;
   count++;
  }

/* dprintf(0,"get frame=%d  count=%d",fn,count); */

  err=tiffgeterror();
  if(!err && count==fn) err=tiff_loadframe(tif,file,fn);

  TIFFClose(tif);
 }

 return(err);
}




static os_error * loadfilepost(int userhandle)
{
 os_error    * err;

 err=NULL;

/* dprintf(0,"load tiff post"); */

 return(err);
 USE(userhandle);
}






static os_error * savetiffim(char * name,int type,imagestr * saveimage)
{
 os_error * err;
 TIFF     * tif;
 float      xres;
 float      yres;
 int        xpix;
 int        ypix;
 int        bpp;
 int        x;
 int        y;
 unsigned short bitspersample;
 unsigned short samplesperpixel;
 int       * idata;
 char      * rowbuff;
 char      * p;
 char      * q;
 int         bytewidth;
 int         compression;
 int         bwimage;
 int         w;

 err=NULL;

 tiffclearerror();

 xres=saveimage->xdpi;
 yres=saveimage->ydpi;
 ypix=saveimage->ypix;
 xpix=saveimage->xpix;
 bpp=saveimage->bpp;
 compression=colour;
 bwimage=imbw(saveimage);


 if(bpp<=8) bytewidth=(xpix*bpp+7)/8;
 else
 {
  if(saveimage->ipal.palclass==CL_CMYK)
  {
   bytewidth=(xpix*4);
  }
  else
  {
   bytewidth=(xpix*3);
  }
 }


 err=salloc((flex_ptr)&rowbuff,bytewidth);
 if(err) return(err);


 tif=TIFFOpen(name,"w");
 if(tif!=NULL)
 {
  TIFFSetField(tif,TIFFTAG_RESOLUTIONUNIT,(unsigned short)RESUNIT_INCH);
  TIFFSetField(tif,TIFFTAG_XRESOLUTION,xres);
  TIFFSetField(tif,TIFFTAG_YRESOLUTION,yres);

  if(saveimage->bpp<=8)
  {
   bitspersample=saveimage->bpp;
   samplesperpixel=1;

   if(bwimage)
   {
    TIFFSetField(tif,TIFFTAG_PHOTOMETRIC,(unsigned short)
                 (bwimage==-1?PHOTOMETRIC_MINISWHITE:PHOTOMETRIC_MINISBLACK));
    if(bw) compression=bw;
   }
   else
   {
    TIFFSetField(tif,TIFFTAG_PHOTOMETRIC,(unsigned short)PHOTOMETRIC_PALETTE);
   }
  }
  else
  {
   TIFFSetField(tif,TIFFTAG_MAXSAMPLEVALUE,(unsigned short)255);
   TIFFSetField(tif,TIFFTAG_MINSAMPLEVALUE,(unsigned short)0);

   if(saveimage->ipal.palclass==CL_CMYK)
   {
    bitspersample=8;
    samplesperpixel=4;
    TIFFSetField(tif,TIFFTAG_PHOTOMETRIC,(unsigned short)PHOTOMETRIC_SEPARATED);
   }
   else
   {
    bitspersample=8;
    samplesperpixel=3;
    TIFFSetField(tif,TIFFTAG_PHOTOMETRIC,(unsigned short)PHOTOMETRIC_RGB);
   }
  }

  TIFFSetField(tif,TIFFTAG_IMAGEWIDTH,saveimage->xpix);
  TIFFSetField(tif,TIFFTAG_IMAGELENGTH,saveimage->ypix);
  TIFFSetField(tif,TIFFTAG_BITSPERSAMPLE,bitspersample);
  TIFFSetField(tif,TIFFTAG_SAMPLESPERPIXEL,samplesperpixel);
  TIFFSetField(tif,TIFFTAG_PLANARCONFIG,(unsigned short)PLANARCONFIG_CONTIG);

  TIFFSetField(tif,TIFFTAG_ROWSPERSTRIP,(unsigned long)8);
  TIFFSetField(tif,TIFFTAG_COMPRESSION,(unsigned short)compression);


  if(saveimage->bpp<=8 && !bwimage)
  {
   unsigned short r[256];
   unsigned short g[256];
   unsigned short b[256];
   int            i;

   for(i=0;i<(1<<saveimage->bpp);i++)
   {
    r[i]=((saveimage->ipal.word[i]>> 8)&0xFF)*257;
    g[i]=((saveimage->ipal.word[i]>>16)&0xFF)*257;
    b[i]=((saveimage->ipal.word[i]>>24)&0xFF)*257;
   }

   TIFFSetField(tif,TIFFTAG_COLORMAP,r,g,b);
  }

  for(y=0;y<ypix;y++)
  {
   err=imfind1r(saveimage,y,&idata);
   if(err) break;

   q=(char*)idata;
   p=rowbuff;

   if(bpp==1) for(x=0;x<bytewidth;x++) *p++=bitrev[*q++];
   else
   if(bpp==2) for(x=0;x<bytewidth;x++) *p++=dbitrev[*q++];
   else
   if(bpp==4) for(x=0;x<bytewidth;x++) *p++=nybrev[*q++];
   else
   if(bpp==8) for(x=0;x<bytewidth;x++) *p++=*q++;
   else
   if(bpp==16)
   {
    for(x=0;x<xpix;x++)
    {
     w=*idata++;

     *p++=(char)((w<<3)&0xF8);
     *p++=(char)((w>>2)&0xF8);
     *p++=(char)((w>>7)&0xF8);

     x++;
     if(x>=xpix) break;

     *p++=(char)((w>>13)&0xF8);
     *p++=(char)((w>>18)&0xF8);
     *p++=(char)((w>>23)&0xF8);
    }
   }
   else
   if(bpp==24)
   {
    for(x=0;x<bytewidth;x++) *p++=*q++;
   }
   else
   if(bpp==32) 
   {
    for(x=0;x<xpix;x++)
    {
     *p++=*q++;
     *p++=*q++;
     *p++=*q++;
     q++;
    }
   }

   if(TIFFWriteScanline(tif,(unsigned char*)rowbuff,y,0)==-1) break;
  }

  TIFFClose(tif);
 }

 if(!err) err=tiffgeterror();
// if(!err) err=fs_settype(name,type);

 sfree((flex_ptr)&rowbuff);

 return(err);
}






os_error * savetiff(filestr * file,int fn,char * name,int type)
{
 imagestr * saveimage;

 saveimage=((file->frames[fn].xim)->sim[IM]);

 return(savetiffim(name,type,saveimage));
}

